export default {

    _meta: {
        label: 'Page Link / Url',
        info: 'Url or page link',
        icon: 'pages:assets/icons/link.svg',
        settings: [
            {name: 'link', type: 'text'},
        ],
        render(value, field, context) {
            return value;
        }
    },

    data() {
        return {
            page: null,
            asset: null,
            val: this.modelValue
        }
    },

    props: {
        modelValue: {
            type: String,
            default: ''
        },
        placeholder: {
            type: String,
            default: 'https://...'
        },
        locale: {
            type: String,
            default: null
        }
    },

    mounted() {

        if (this.val && this.val.match(/^pages:\/\//)) {
            this.loadPage();
        }

        if (this.val && this.val.match(/^assets:\/\//)) {
            this.loadAsset();
        }
    },

    computed: {
        hasLinkedPage() {
            return this.page && this.val && this.val.indexOf(this.page._id) !== -1;
        },

        hasLinkedAsset() {
            return this.asset && this.val && this.val.indexOf(this.asset._id) !== -1;
        },

        assetPreview() {

            if (!this.asset) {
                return null;
            }

            let mediaSize = 50;
            let media = `<kiss-svg width="${mediaSize}" height="${mediaSize}" src="${App.base(this.asset.type === 'video' ? 'assets:assets/icons/video.svg' : 'assets:assets/icons/file.svg')}"><canvas width="${mediaSize}" height="${mediaSize}"></canvas></kiss-svg>`;

            if (this.asset.type === 'image') {

                const src = App.route(`/assets/thumbnail/${this.asset._id}?m=bestFit&mime=auto&h=${(mediaSize * 2)}&t=${this.asset._modified}`);

                media = `<img loading="lazy" class="kiss-responsive-height kiss-margin-auto" src="${src}" title="${this.asset.title}" style="height:${mediaSize}px;object-fit:contain;object-position:center;">`;
            }

            return media;

        }
    },

    watch: {
        modelValue() {
            this.val = this.modelValue;
            this.update();
        }
    },

    methods: {

        selectPage() {

            VueView.ui.offcanvas('pages:assets/dialogs/page-picker.js', {page: this.val, locale: this.locale}, {

                select: (page, locale) => {
                    this.page = page;
                    this.val = `pages://${page._id}`+((locale && locale != 'default' && `?locale=${locale}`) || '');
                    this.update();
                }

            }, {flip: true, size: 'large'})

        },

        selectAsset() {

            App.utils.selectAsset(asset => {
                this.asset = asset;
                this.val = `assets://${asset._id}`
                this.update();
            });
        },

        loadPage() {

            if (!this.val || !this.val.match(/^pages:\/\//)) {
                return;
            }

            const url = new URL(this.val), locale = url.searchParams.get('locale');

            let options = {
                filter: {_id: url.host},
                limit: 1
            };

            if (locale && App._locales[locale]) {
                options.locale = locale;
            }

            this.$request('/pages/utils/find', {options}).then(rsp => {

                this.page = null;

                if (rsp.items && rsp.items[0]) {
                    this.page = rsp.items[0];
                }
            }).catch(res => {
                App.ui.notify(res.error || 'Loading page failed!', 'error');
            });
        },

        loadAsset() {

            if (!this.val || !this.val.match(/^assets:\/\//)) {
                return;
            }

            const url = new URL(this.val);

            let options = {
                filter: [{_id: url.host}],
                limit: 1
            };

            this.$request('/assets/assets', { options }).then(rsp => {
                this.asset = rsp.assets[0] ?? null;
            });

        },

        update() {
            this.$emit('update:modelValue', this.val)
        }
    },

    template: /*html*/`
        <div field="page-url">

            <kiss-card class="kiss-padding-small kiss-size-small kiss-position-relative kiss-margin-small" v-if="hasLinkedPage" theme="bordered contrast" :style="{borderColor: page._state !== 1 ? 'var(--kiss-color-danger)':''}">
                <div class="kiss-text-bold">{{ page.title }}</div>
                <div class="kiss-color-muted kiss-size-xsmall kiss-text-truncate">{{page._r}}</div>
                <a class="kiss-cover" :href="$routeUrl('/pages/page/'+page._id)" target="_blank"></a>
            </kiss-card>

            <kiss-card class="kiss-size-small kiss-position-relative kiss-margin-small" v-if="hasLinkedAsset" theme="bordered contrast">
                <kiss-row class="kiss-flex-middle" gap="small">
                    <div class="kiss-position-relative">
                        <canvas width="50" height="50"></canvas>
                        <div class="kiss-cover" :class="{'kiss-bgcolor-transparentimage': asset.type === 'image'}" style="--kiss-bgcolor-transparentimage-size:10px" v-html="assetPreview"></div>
                    </div>
                    <div class="kiss-flex-1 kiss-size-xsmall">
                        <div class="kiss-text-truncate" :title="asset.title">{{asset.title}}</div>
                        <div class="kiss-color-muted">{{ App.utils.formatSize(asset.size) }}</div>
                    </div>
                </kiss-row>
                <a class="kiss-cover" :href="$baseUrl('#uploads:'+asset.path)" target="_blank"></a>
            </kiss-card>

            <kiss-navlist class="kiss-flex">
                <input type="text" class="kiss-input kiss-flex-1 kiss-margin-small-right" :class="{'kiss-color-muted': hasLinkedPage}" v-model="val" @input="update" :placeholder="placeholder">
                <kiss-dropdown>
                    <button class="kiss-button" type="button">Pick</button>
                    <kiss-dropdownbox>
                        <kiss-navlist>
                            <ul>
                                <li class="kiss-nav-header">{{ t('Resource') }}</li>
                                <li><a @click="selectPage"><icon>link</icon> {{ t('Page') }}</a></li>
                                <li><a @click="selectAsset"><icon>attach_file</icon> {{ t('Asset') }}</a></li>
                            </ul>
                        </kiss-navlist>
                    </kiss-dropdownbox>
                </kiss-dropdown>
            </div>
        </div>
    `,
}
